/* rc.c
 * run clm note list (with-sound) as a separate C program.
 * 
 * this is for those implementations of Lisp that do not have an adequate foreign function interface (ffi)
 * and is activated in CLM (all.lisp) by pushing :no-ffi on the *features* list.
 *
 */

#include <config.h>
#include <math.h>
#include <stdio.h>
#if HAVE_FCNTL_H
  #include <fcntl.h>
#endif
#if HAVE_LIMITS_H
  #include <limits.h>
#endif
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#if defined(SGI) || defined(LINUX) || defined(SUN) || defined(ALPHA)
  #include <sys/time.h>
#endif

#if (defined(HAVE_LIBC_H) && (!defined(HAVE_UNISTD_H)))
  #include <libc.h>
#else
  #ifndef _MSC_VER
    #include <unistd.h>
  #endif
#endif

#ifdef SGI
#include <sys/fpu.h>
#endif

#ifndef CLM_FILE_BUFFER_SIZE
  #define CLM_FILE_BUFFER_SIZE 65536
#endif


void clm_c_break(void) {}
void clm_c_error(void) {}

#define FASMIX -2
#define REVERB -1

/* from initmus.lisp */
#define CLM_INTERRUPTED 0
#define CURRENT_OUTPUT_FILE 2
#define REVERB_FILE 3

static const unsigned char I_dcdt[4] = {'.','c','d','t'};

/* timing info */
#if defined(LINUX) || defined(SGI) || defined(SUN) || defined(ALPHA)
  static struct timeval t0,t1;
  static struct timezone z0,z1;
#else
  static clock_t c1,c2;
#endif

static void start_time (void)
{
#if defined(LINUX) || defined(SUN) || defined(ALPHA)
  gettimeofday(&t0,&z0);
#else
  #if defined(SGI)
    gettimeofday(&t0);
  #else
    #if defined(MACOS) || defined(WINDOZE)
      c1=clock();
    #endif
  #endif
#endif
}

static int run_time (void)
{
  int secs,millisecs;
#if defined(LINUX) || defined(SUN) || defined(ALPHA)
  gettimeofday(&t1,&z1);
#else
  #if defined(SGI)
    gettimeofday(&t1);
  #else
    #if defined(MACOS) || defined(WINDOZE)
      c2=clock();
      return((int)(1000.0*((float)(c2-c1))/(float)CLOCKS_PER_SEC));
    #endif
  #endif
#endif
#if defined(LINUX) || defined(SUN) || defined(ALPHA) || defined(SGI)
  secs = (t1.tv_sec - t0.tv_sec);
  millisecs = ((t1.tv_usec - t0.tv_usec) / 1000);
  return(secs*1000+millisecs);
#endif
}

typedef int (*intinsfun)();
static intinsfun *ins_pool;        /* entry points of instruments */

#include "rc.pool"

static int match_four_chars(unsigned char *head, const unsigned char *match)
{ 
  int i;
  for (i=0;i<4;i++) if (head[i] != match[i]) return(0);
  return(1);
}
  
static int as_int(int fd, unsigned char *buf)
{
  read(fd,buf,4);
  return(mus_char_to_bint(buf));
}

static float as_float(int fd, unsigned char *buf)
{
  read(fd,buf,4);
  return(mus_char_to_bfloat(buf));
}

static int last_string_len;

static char *as_string(int fd, unsigned char *buf)
{
  int len;
  char *str;
  len=as_int(fd,buf);
  last_string_len=len;
  str = (char *)calloc(len+1,sizeof(char));
  read(fd,str,len);
  return(str);
}

static int last_int_len;

static int *as_int_array(int fd, unsigned char *buf, int *data, int *data_len)
{
  int len;
#if MUS_LITTLE_ENDIAN
  int i;
  unsigned char *dp;
#endif
  len = as_int(fd,buf);
  last_int_len = len;
  if (len > (*data_len))
    {
      if ((*data_len) != 0) free(data);
      (*data_len) = (2*len);
      data = (int *)calloc((*data_len),sizeof(int));
    }
  read(fd,(char *)data,len*4);
#if MUS_LITTLE_ENDIAN
  for (i=0,dp=(unsigned char *)data;i<len;i++,dp+=4) data[i]=mus_char_to_bint(dp);
#endif
  return(data);
}

static float *as_float_array(int fd, unsigned char *buf, float *data, int *data_len, int *cur_len)
{
  int len;
#if MUS_LITTLE_ENDIAN
  int i;
  unsigned char *dp;
#endif
  len = as_int(fd,buf);
  (*cur_len) = len;
  if (len > (*data_len))
    {
      if ((*data_len) != 0) free(data);
      (*data_len) = (2*len);
      data = (float *)calloc((*data_len),sizeof(float));
    }
  read(fd,(char *)data,len*4);
#if MUS_LITTLE_ENDIAN
  for (i=0,dp=(unsigned char *)data;i<len;i++,dp+=4) data[i]=mus_char_to_bfloat(dp);
#endif
  return(data);
}

static void update_io(int *outb, int *datain, int adr)
{
  outb[CLM_IO_DATA_END] = datain[CLM_IO_DATA_END+adr];
  outb[CLM_IO_DATA_START] = datain[CLM_IO_DATA_START+adr];
  outb[CLM_IO_SIZE] = datain[CLM_IO_SIZE+adr];
  outb[CLM_IO_BEG] = datain[CLM_IO_BEG+adr];
  outb[CLM_IO_END] = datain[CLM_IO_END+adr];
}

static int *initialize_io(int outf_fd, int ochans, int ohdrend)
{
  int *outf;
  int i;
  outf = (int *)calloc(CLM_IO_DATS+ochans+6,sizeof(int));
  outf[0] = 37;
  outf[CLM_IO_FD] = outf_fd;                        /* set up CLM's view of the output file */
  outf[CLM_IO_CHANS] = ochans;
  outf[CLM_IO_SIZE]=0;
  outf[CLM_IO_BEG]=0;
  outf[CLM_IO_END]=CLM_FILE_BUFFER_SIZE-1;
  outf[CLM_IO_BUFSIZ]=CLM_FILE_BUFFER_SIZE;
  outf[CLM_IO_DATA_START]=0;
  outf[CLM_IO_DATA_END]=0;
  outf[CLM_IO_DIR]=CLM_IO_OUT_FILE;
  outf[CLM_IO_HDR_END]=ohdrend;
  outf[CLM_IO_DATS+CLM_AREF_IBLOCK]=CLM_IO_DATS+6;
  outf[CLM_IO_DATS+CLM_AREF_TYPE]=39; /* +integer-array+ */
  outf[CLM_IO_DATS+CLM_AREF_SIZE]=ochans;
  outf[CLM_IO_DATS+CLM_AREF_DIMS]=1;
  for (i=0;i<ochans;i++) outf[CLM_IO_DATS+6+i]=(int)MUS_MAKE_SAMPLE_ARRAY(CLM_FILE_BUFFER_SIZE);
  return(outf);
}

static int open_files[32];
static mus_sample_t *open_file_buffers[256];
static int open_file_ctr = 0;
static int open_file_buffer_ctr = 0;
int rc_open(char *ifile, int *datai, int loc);

int rc_open(char *ifile, int *datai, int loc)
{
  int ifd,ichans,i;
  if ((ifd=mus_file_open_read(ifile)) == -1) return(-1);
  ichans = mus_sound_chans(ifile);
  mus_file_open_descriptors(ifd,ifile,
			    mus_sound_data_format(ifile),mus_sound_datum_size(ifile),mus_sound_data_location(ifile),
			    ichans,mus_sound_header_type(ifile));
  datai[loc+CLM_IO_DATS+CLM_AREF_IBLOCK]=loc+CLM_IO_DATS+6;
  datai[loc+CLM_IO_DATS+CLM_AREF_TYPE]=39;
  datai[loc+CLM_IO_DATS+CLM_AREF_SIZE]=ichans;
  datai[loc+CLM_IO_DATS+CLM_AREF_DIMS]=1;
  datai[loc+CLM_IO_END] = -1;
  for (i=0;i<ichans;i++) 
    {
      open_file_buffers[open_file_buffer_ctr] = MUS_MAKE_SAMPLE_ARRAY(CLM_FILE_BUFFER_SIZE);
      datai[loc+CLM_IO_DATS+6+i] = (int)(open_file_buffers[open_file_buffer_ctr]);
      open_file_buffer_ctr++;
    }
  open_files[open_file_ctr++] = ifd;
  return(ifd);
}

#define CDT_BUFFER_SIZE 65536

int main (int argc, char *argv[])
{
  /* read .cdt file, producing CLM output sound.
   *
   * argv[1] = <mumble.cdt> (our current data file)
   *
   * ----------------
   * ".cdt"
   * srate
   * scaled-to value (needed right away to fix up output types)
   * output file name length (0 if direct to dac?)
   *   output file name
   *   output channels
   *   output data format,,type
   *   output type
   *   output header end
   * output header length
   *   output header
   * reverb file name length (0 if none)
   *   reverb file name
   *   reverb channels
   *   reverb data format,,type
   *   reverb type
   *   reverb header end
   * reverb header length
   *   reverb header
   * stats flag
   * final data format (for scaled-to)
   * blocks describing instrument calls, each call:
   *   index into instrument function array (ins_pool in rc.pool)
   *   beg
   *   end
   *   datar length (floats)
   *   datar
   *   datai length (ints)
   *   datai
   * EOF or -1 = signal that next block is reverb call
   *   possible reverb instrument call (just like above)
   * ----------------
   *
   * each with-sound call, if rc is in effect, opens the .cdt file, sets up the header info,
   *   then on each instrument call, we write the data to the .cdt array rather than call the function.
   *   When all the data is out (at end of with-sound), we call rc with the .cdt name.
   * We also create upon with-sound starting up a list of currently active instruments from
   *   *clm-instruments* which has the associated C file and function names.  This file is
   *   written/read alongside the .cdt file.
   */

  int *outf,*revf;
  int *datai;
  mus_sample_t *o_vals,*r_vals;
  off_t *o_times, *r_times;
  float *datar;
  double *ins_datar;
  int ins_datar_len,idl;
  float scaled_to,maxamp,rtime,amp;
  char *cdt_filename = NULL,*outf_filename = NULL,*revf_filename = NULL;
  char *fas_infile = NULL,*fas_outfile = NULL,*old_outf_filename = NULL;
  char *final_comment = NULL;
  unsigned char *cdt_buffer;
  int cdt_fd,outf_fd,revf_fd,ochans,rchans,i,j,odataformat,rdataformat;
  int otype,rtype,len,srate,stats,bytes,ohdrend,rhdrend,ofilenamelen,comment_len;
  int beg,end,p_index,happy,cur_datar_len,cur_datai_len,outf_adr,revf_adr;
  int inbeg,outbeg,indur,samples,final_format,last_datai_len,oclipped;

#ifdef SGI
  union fpc_csr f;
  f.fc_word = get_fpc_csr();
  f.fc_struct.flush = 1;
  set_fpc_csr(f.fc_word);
#endif

  cur_datar_len = 0;
  cur_datai_len = 0;
  revf_fd = 0;

  mus_sound_initialize();
  initialize_cmus();

  ins_pool_initialize();

#ifndef MACOS
  cdt_filename=argv[1];  /* we're called rc <file>.cdt */
  #ifdef WINDOZE
    cdt_fd = open (cdt_filename, O_RDONLY | O_BINARY);
  #else
    cdt_fd = open(cdt_filename,O_RDONLY,0);
  #endif
#else
  cdt_filename = "test.cdt";
  cdt_fd = open(cdt_filename,O_RDONLY);
#endif
  if (cdt_fd == -1) {printf("can't open clm no-ffi data file: %s",cdt_filename); exit(1);}
  cdt_buffer = (unsigned char *)calloc(CDT_BUFFER_SIZE,sizeof(unsigned char));
#ifndef MACOS
  read(cdt_fd,cdt_buffer,4);
#else
  read(cdt_fd,(char *)cdt_buffer,4);
#endif
  if (!(match_four_chars((unsigned char *)cdt_buffer,I_dcdt))) {printf("%s does not start with .cdt!",cdt_filename); exit(1);}

  srate = as_int(cdt_fd,cdt_buffer);
  scaled_to = as_float(cdt_fd,cdt_buffer);
  outf_filename = as_string(cdt_fd,cdt_buffer);
  ofilenamelen=last_string_len;
  ochans = as_int(cdt_fd,cdt_buffer);
  odataformat = as_int(cdt_fd,cdt_buffer);
  oclipped = odataformat >> 16;
  odataformat = (odataformat & 0xffff);
  otype = as_int(cdt_fd,cdt_buffer);
  if ((scaled_to != 0.0) && (otype == MUS_RIFF)) {otype = MUS_NEXT; odataformat = MUS_BINT;}
  ohdrend = as_int(cdt_fd,cdt_buffer);
  final_comment = as_string(cdt_fd,cdt_buffer);
  comment_len = last_string_len;
  outf_fd = mus_sound_open_output(outf_filename,srate,ochans,odataformat,otype,final_comment);
  lseek(outf_fd,0,SEEK_SET);
  ohdrend=mus_sound_data_location(outf_filename);
  outf = initialize_io(outf_fd,ochans,ohdrend);

  if (oclipped==1) mus_file_set_data_clipped(outf_fd,1);

  len = as_int(cdt_fd,cdt_buffer);              /* reverb file name length (0=no reverb) */
  if (len > 0)
    {
      revf_filename = (char *)calloc(len,sizeof(char));
      read(cdt_fd,revf_filename,len);   
      rchans = as_int(cdt_fd,cdt_buffer);
      rdataformat = as_int(cdt_fd,cdt_buffer);
      rtype = as_int(cdt_fd,cdt_buffer);
      if ((scaled_to != 0.0) && (rtype == MUS_RIFF)) {rtype = MUS_NEXT; rdataformat = MUS_BINT;}
      rhdrend = as_int(cdt_fd,cdt_buffer);
      len = as_int(cdt_fd,cdt_buffer);
#if (!(defined(MACOS))) && (!(defined(WINDOZE)))
      read(cdt_fd,cdt_buffer,len);
      revf_fd = mus_sound_open_output(revf_filename,srate,rchans,rdataformat,rtype,cdt_buffer);
#else
      read(cdt_fd,(char *)cdt_buffer,len);
      revf_fd = mus_sound_open_output(revf_filename,srate,rchans,rdataformat,rtype,(char *)cdt_buffer);
#endif
      lseek(revf_fd,0,SEEK_SET);
      rhdrend=mus_sound_data_location(revf_filename);
      revf = initialize_io(revf_fd,rchans,rhdrend);
      if (oclipped) mus_file_set_data_clipped(revf_fd,1);
    }
  stats = as_int(cdt_fd,cdt_buffer);
  final_format = as_int(cdt_fd,cdt_buffer);
  if (stats) start_time();

  happy = 1;
  while (happy)
    {
#ifndef MACOS
      bytes = read(cdt_fd,cdt_buffer,4);        /* p_index */
#else
      bytes = read(cdt_fd,(char *)cdt_buffer,4);
#endif
      if (bytes < 4) happy = 0;                 /* EOF = all done */
      else
	{
	  p_index = mus_char_to_bint(cdt_buffer);
	  if (p_index == FASMIX)
	    {
	      clm_file_reset(0,outf,outf); /* flush-snd */
	      fas_infile = as_string(cdt_fd,cdt_buffer);
	      fas_outfile = as_string(cdt_fd,cdt_buffer);
	      outbeg = as_int(cdt_fd,cdt_buffer);
	      inbeg = as_int(cdt_fd,cdt_buffer);
	      indur = as_int(cdt_fd,cdt_buffer);
	      amp = as_float(cdt_fd,cdt_buffer);
	      samples = clm_simple_mix(fas_outfile, fas_infile, outbeg, indur, inbeg, (double)amp, 1.0);
	      len = samples+outbeg;
	      if (outf[CLM_IO_SIZE] < len) outf[CLM_IO_SIZE] = len;
	    }
	  else
	    {
	      if (p_index == REVERB)
		{
		  /* slightly different from a normal with-sound call because we can't treat reverb as the local
		   * input file at the lisp call -- we need to kludge the pointers at this level (as a local input
		   * file, not as reverb).
		   */
		  clm_file_reset(0,revf,revf);
		  mus_sound_close_output(revf_fd,mus_samples_to_bytes(rdataformat,revf[CLM_IO_SIZE]));
		  revf_fd = mus_sound_open_input(revf_filename);
		  revf[CLM_IO_FD]=revf_fd;
		  revf[CLM_IO_DIR]=CLM_IO_IN_FILE;
		  revf[CLM_IO_BEG]=0;
		  revf[CLM_IO_END]=-1;
		  revf[CLM_IO_DATA_START]=0;
		  revf[CLM_IO_DATA_END]=0;
		  p_index = as_int(cdt_fd,cdt_buffer);
		}
		  
	      beg = as_int(cdt_fd,cdt_buffer);
	      end = as_int(cdt_fd,cdt_buffer);
	      datar = as_float_array(cdt_fd,cdt_buffer,datar,&cur_datar_len,&ins_datar_len);
	      ins_datar = (double *)calloc(ins_datar_len,sizeof(double));
	      for (idl=0;idl<ins_datar_len;idl++) ins_datar[idl] = (double)(datar[idl]);

	      datai = as_int_array(cdt_fd,cdt_buffer,datai,&cur_datai_len);
	      last_datai_len = last_int_len;
	      
	      outf_adr = datai[CURRENT_OUTPUT_FILE];
	      if (outf_adr != 0)
		{
		  for (i=outf_adr,j=0;i<(outf_adr+CLM_IO_DATS+6+ochans);i++,j++) datai[i]=outf[j];
		  datai[outf_adr+CLM_IO_DATS+CLM_AREF_IBLOCK] = outf_adr+CLM_IO_DATS+6;
		}

	      if (revf_fd)
		{
		  revf_adr = datai[REVERB_FILE];
		  for (i=revf_adr,j=0;i<(revf_adr+CLM_IO_DATS+6+rchans);i++,j++) datai[i]=revf[j];
		  datai[revf_adr+CLM_IO_DATS+CLM_AREF_IBLOCK]=revf_adr+CLM_IO_DATS+6;
		}
	      open_file_ctr = 0;
	      (*ins_pool[p_index])(beg,end,ins_datar,0,datai,0); 
	      if (open_file_buffer_ctr > 0)
		{
		  for (i=0;i<open_file_buffer_ctr;i++)
		    if (open_file_buffers[i])
		      {
			free(open_file_buffers[i]);
			open_file_buffers[i] = NULL;
		      }
		  open_file_buffer_ctr = 0;
		}
	      free(ins_datar);
	      if (open_file_ctr > 0)
		{
		  for (i=0;i<open_file_ctr;i++) mus_file_close(open_files[i]);
		}
	      update_io(outf,datai,outf_adr);
	      if (revf_fd && (revf[CLM_IO_DIR] == CLM_IO_OUT_FILE)) update_io(revf,datai,revf_adr);
	    }
	}
    }

  clm_file_reset(0,outf,outf);
  samples = outf[CLM_IO_SIZE];
  mus_sound_close_output(outf_fd,outf[CLM_IO_CHANS]*mus_samples_to_bytes(odataformat,outf[CLM_IO_SIZE]));
  if (revf_fd) mus_file_close(revf_fd);
  close(cdt_fd);

  if ((stats != 0) || (scaled_to != 0.0)) /* either stats flag or scaled-to? */
    {
      o_vals = (mus_sample_t *)calloc(ochans,sizeof(mus_sample_t));
      o_times = (off_t *)calloc(ochans, sizeof(off_t));
      mus_sound_maxamps(outf_filename,ochans,o_vals,o_times);
      maxamp = MUS_SAMPLE_TO_FLOAT(o_vals[0]);
      for (i=1;i<ochans;i++) if (MUS_SAMPLE_TO_FLOAT(o_vals[i]) > maxamp) maxamp = MUS_SAMPLE_TO_FLOAT(o_vals[i]);
      if (revf_fd)
	{
	  r_vals = (mus_sample_t *)calloc(1,sizeof(mus_sample_t));
	  r_times = (off_t *)calloc(1, sizeof(off_t));
	  mus_sound_maxamps(revf_filename,1,r_vals, r_times);
	  if (MUS_SAMPLE_TO_FLOAT(r_vals[0]) > maxamp) maxamp = MUS_SAMPLE_TO_FLOAT(r_vals[0]);
	}
    }

  if (stats)
    {
      printf("%s: Duration: %.4f\n",outf_filename,(float)samples/(float)srate);
      if (samples != 0)
	{
	  rtime = (float)run_time()/1000.0;
	  printf("  Compute time: %.3f, Compute ratio: %.2f\n",rtime,(rtime*((float)srate/(float)samples)));
	}
      printf("  OutA max amp: %.3f (near %.3f secs)\n",MUS_SAMPLE_TO_FLOAT(o_vals[0]),(float)o_times[0]/(float)srate);
      if (ochans > 1)
	{
	  printf("  OutB max amp: %.3f (near %.3f secs)\n",MUS_SAMPLE_TO_FLOAT(o_vals[1]),(float)o_times[1]/(float)srate);
	  if (ochans > 2)
	    {
	      printf("  OutC max amp: %.3f (near %.3f secs)\n",MUS_SAMPLE_TO_FLOAT(o_vals[2]),(float)o_times[2]/(float)srate);
	      if (ochans > 3)
		{
		  printf("  OutD max amp: %.3f (near %.3f secs)\n",MUS_SAMPLE_TO_FLOAT(o_vals[3]),(float)o_times[3]/(float)srate);
		  if (ochans > 4)
		    {
		      for (i=4;i<ochans;i++)
			{
			  printf("  Out%d max amp: %.3f (near %.3f secs)\n",i,MUS_SAMPLE_TO_FLOAT(o_vals[i]),(float)o_times[i]/(float)srate);
			}
		    }
		}
	    }
	}
      if (revf_fd)
	{
	  printf("  RevA max amp: %.3f (near %.3f secs)\n",MUS_SAMPLE_TO_FLOAT(r_vals[0]),(float)r_times[0]/(float)srate);
	}
    }

  if (scaled_to != 0.0)
    {
      if (maxamp == 0.0) maxamp=1.0;
      old_outf_filename = outf_filename; /* chop off the .temp */
      ofilenamelen = strlen(outf_filename);
      outf_filename = (char *)calloc(ofilenamelen,sizeof(char));
      for (i=0;i<ofilenamelen-5;i++) outf_filename[i]=old_outf_filename[i];
      if (final_format == MUS_LSHORT) otype = MUS_RIFF;
      clm_scale_file(outf_filename,old_outf_filename,scaled_to/maxamp,final_format,otype);
      remove(old_outf_filename);
    }
  return(0);
}
